'
' -----[ Title ]------------------------------------------------------
'
' File......: SNAP-002.BAS
' Purpose...: Lampdimmer node for PLM-24
' Author....: Robert Zachrisson
' Version...: 1.02
' Stamp.....: BS1-IC
' Started...: 980501
' Updated...: 990125
'
' -----[ Program Description ]----------------------------------------
'
' This program shows how to implement the S.N.A.P protocol in a BS1-IC
' and use it to dim a lamp. It uses PLM-24 Power Line Modems to 
' communicate over the power lines (mains). This example uses 16-bit 
' CRC-CCITT as error detection method which gives secure data transfer.
'
' If this node is addressed by another node (PC or another MCU) the data
' byte (DB1) will contain the dimmer level value, used to dim the lamp.
' The higher the value the stronger the lamp will lit
'
' We use the watchdog in the PLM-24 to reset the BS1-IC if it only gets
' part of a packet and therefor get stuck in the serin command.
' Or if it hangs for any other reason. The BS1-IC makes a beep when it
' starts, so every time it misses a packet you will hear that.
'
' Since the BS1-IC lack interrupt capabilities we use a triac curcuit
' manufactured by United Automation Ltd. in UK model 1004A (rated for
' 230VAC, 10 Amp). This dimmer module is controlled by a variable 
' resistor and in our case  we use an LED -> LDR opto coupler made by
' Vactec model VTL5C3.
'
' The packet structure is defined in the received packets first two
' bytes (HDB2 and HDB1). The following packet structure is used.
'
' DD=01     - 1 Byte destination address
' SS=01     - 1 Byte source address
' PP=00     - No protocol specific flags
' AA=01     - Acknowledge is required
' C=0       - Command mode not supported
' EEE=100   - 16-bit CRC-CCITT
' NNNN=0001 - 1 Byte data
'
' Overview of header definition bytes (HDB2 and HDB1)
'
'        HDB2              HDB1
' +-----------------+-----------------+
' | D D S S P P A A | C E E E N N N N |
' +-----------------+-----------------+
'
' For more information about PLM-24 Power Line Modem see URL below.
'
' http://www.hth.com/plm-24/
'
' For more information about S.N.A.P see URL below.
'
' http://www.hth.com/snap/
'
' Below is an short overview of the S.N.A.P network protocol.
'
' Features
' --------
'
' - Free and open network protocol.
' - Easy to learn, use and implement.
' - Up to 16.7 million destination node addresses.
' - Up to 16.7 million source node addresses.
' - 0-24 protocol specific flags.
' - Optional ACK/NAK request.
' - Optional command mode.
' - 8 different error detecting methods.
' - Header is scaleable from 3-14 bytes.
' - 0-512 Bytes data in each packet.
' - Can be used as master/slave or peer-to-peer.
' - Supports broadcast messages.
'
' -----[ Disclaimer ]-------------------------------------------------
' 
' This example is offered on an "AS IS" basis, no warranty expressed
' or implied. HTH disclaim liability of any damages associated with
' the use of the hardware or software described herein. You use it on
' your own risk. We are not able to provide any free support.
'
' Copyright (c) 1998-1999 HTH all rights reserved
'
' -----[ Revision History ]-------------------------------------------
'
' 980505 - Version 1.00    Initial release
' 980911 - Version 1.01 Revised for S.N.A.P Version 0.91
' 990125 - Version 1.02 Revised for S.N.A.P Version 1.00
'
' -----[ I/O Pin usage ]----------------------------------------------
' 
' Stamp pinNr/name	In/Out		External hardware	 
'
' Pin 14/IO7		In		Not connected		
' Pin 13/IO6	  	   Out		Optocoupler LED
' Pin 12/IO5		   Out          Piezospeaker
' Pin 11/IO4		   Out		Plm-24 WD pin
' Pin 10/IO3	  	   Out 		PLM-24 Rx/Tx pin
' Pin  9/IO2		In		PLM-24 CD pin
' Pin  8/IO1	   	   Out		PLM-24 TxD pin 
' Pin  7/IO0	 	In		PLM-24 RxD pin
' Pin  6/reset		In		Inverter, then to PLM-24 RSTO pin

' -----[ Constants ]--------------------------------------------------
'
symbol Preamble		= %01010101	' Preamble
symbol SYNC		= %01010100	' Syncronisation byte	
symbol RxD		= 0             ' PLM-24 Serial output pin
symbol TxD		= 1             ' PLM-24 Serial input pin
symbol CarrierDetectInv	= pin2          ' PLM-24 Carrier detect pin (active low)
symbol RxTx		= 3             ' PLM-24 Rx/Tx control pin
symbol WatchDogInv	= 4		' PLM-24 Watch Dog pin (active low)
symbol Piezospeaker	= 5
symbol LED		= 6		' Dimmer opto coupler LED pin
symbol Baud		= T2400         ' Baudrate
symbol CRCPOLY		= $1021         ' CRC-CCITT
symbol MyAddress	= $02           ' Address for this node (1-255)
 		    
' -----[ Variables ]--------------------------------------------------
'
symbol CRC		= w0	' CRC Word
symbol Lo_CRC		= b0	' CRC Lo_Byte
symbol Hi_CRC		= b1	' CRC Hi_Byte
symbol DB1		= b2	' Packet data
symbol SAB1		= b3	' What node sent this packet
symbol DAB1		= b4	' What node should have this paket
symbol HDB1		= b5	' Header Definition Byte 1
symbol HDB2		= b6	' Header Definition Byte 2
symbol DimmerLevel	= b7	' Value that controls the dimmer level
symbol Byte		= b8	' Temporary Variable
symbol Temp		= b9	' Temporary Variable
symbol CRC2		= b10	' Packet CRC Hi_Byte
symbol CRC1		= b11   ' Packet CRC Lo_Byte

' -----[ Initialization ]---------------------------------------------
'
High RxTx : High TxD		' Set IOpin 3 and 1 high on startup
Dirs = %00111010		' Set IOpin 5,4,3 and 1 as outputs
Read 0,DimmerLevel	'Get dimmer level stored before lastpower down
Sound Piezospeaker,(110,8,120,6,124,4)	'make wakeup from reset bip

'-------[ Program ]----------------------------------------------------
'
Start:
        'Keep the watchdog from reseting the Stamp
        Pulsout WatchDogInv,1
	
        'Dim the LED, this will dim the lamp
        Pwm LED,DimmerLevel,1
	
        'Check if there is new data from the PLM-24 modem, if not,
        ' continue to dim the Lamp 
        if CarrierDetectInv = 1 then Start

Receive:
        ' Wait for sync byte, if received, read next seven bytes from master
        serin RxD,baud,(SYNC),HDB2,HDB1,DAB1,SAB1,DB1,CRC2,CRC1

PacketHeader_check:
        ' Check HDB2 to see if STAMP are capable to use the packet
        ' structure, if not goto Start
        if HDB2 <> %01010001 then Start
        ' Check HDB1 to see if STAMP are capable to use the packet
        ' structure, if not goto Start
        if HDB1 <> %01000001 then Start

Address_check:
        ' Check if this is the node addressed, if not goto Start
        if DAB1 <> MyAddress then Start

Error_check:
        ' Check if the data is error free
        gosub CRC_check
        ' Continue to check data
        Byte = CRC2 
        gosub calc_CRC
        Byte = CRC1
        gosub calc_CRC 

ACK_NAK:        
        ' Prepare for ACK/NAK
        temp = SAB1      ' Swap address bytes
        SAB1 = DAB1      ' Swap address bytes
        DAB1 = temp      ' Swap address bytes
	
        ' Check if there was any CRC errors, if so send NAK
        if CRC <> 0 then NAK

        ' No CRC errors in packet so change DimmerLevel to new value,
        ' and store it in non volatile EEPROM, so that it can be 
        ' retrieved after power failure or watchdog reset.
        ' Make OK bip, and return an ACK packet to source (to
        ' indicate to sender that the packet it sent, has been received
        ' and was error free) 
        DimmerLevel = DB1
        Write 0,DimmerLevel
        sound Piezospeaker,(125,1)              

ACK:
        ' Send ACK (i.e tell master that packet was OK)
        ' Set ACK bits to ACK in HDB2 (logic XOR bit0 and bit1)
        HDB2 = HDB2 ^ %00000011
        goto send

NAK:
        ' Send NAK (i.e tell master that packet was bad)
        ' Set ACK bits to NAK in HDB2 (logic OR bit0 and bit1)
        HDB2 = HDB2 | %00000011
        goto send

Send:
        'Calculate CRC bytes
        Gosub CRC_check
        'Move calculated Hi_CRC value to outgoing packet
        CRC1 = Hi_CRC
        'Move calculated Lo_CRC value to outgoing packet
        CRC2 = Lo_CRC
        
Check_CD:
        'Check carrier detect pin if some other PLM-24 unit is sending
        if CarrierDetectInv = 1 then Tx

        'Someone else was sendning. Use current CRC value to generate
        'a random pause time between 0-255 ms before trying again
        random CRC
        pause Lo_CRC
        goto Check_CD

Tx:
        'Set PLM-24 Power Line Modem in transmit mode
        low RxTx
        
        'Give PLM-24 time to settle before transmitting any data
        pause 50

        'Send packet to master, including three preamble bytes and
        'the sync byte
        serout TxD,baud,(Preamble,Preamble,SYNC,HDB2,HDB1,DAB1,SAB1,DB1,CRC2,CRC1)

        'Give STAMP time to shift out all bits before setting to Rx
        Pause 50
        
        'Set PLM-24 Power Line Modem in receive mode
        high RxTx

        'Done, go back to Start and continue to dim the lamp
        goto start
 
' -----[ Subroutines ]------------------------------------------------


CRC_check:
        'Calculate CRC for all received bytes
        CRC = 0
        Byte = HDB2
        gosub calc_CRC
        Byte = HDB1
        gosub calc_CRC
        Byte = DAB1
        gosub calc_CRC
        Byte = SAB1
        gosub calc_CRC
        Byte = DB1
        gosub calc_CRC
	return


calc_crc:
        'Subroutine for calculating CRC value in variable byte 
        CRC = Byte * 256 ^ CRC
        for temp = 0 to 7
        if Bit15 = 0 then shift_only
        CRC = CRC * 2 ^ CRCPOLY
        goto nxt
shift_only: CRC = CRC * 2

nxt:
        'Dim the lamp
        Pwm LED,DimmerLevel,1
	next
	return

' -----[ End ]--------------------------------------------------------

